# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 11:54:13 2016

@author: deeplearning.net, edited for regression by OKM
"""

"""
"""
import os
import sys
import timeit
import cPickle

import numpy
import scipy.io

import theano
import theano.tensor as T
from theano.sandbox.rng_mrg import MRG_RandomStreams

from linregClass import LinearRegression
from MLPclass import HiddenLayer
from RBM import RBM
from GBRBM import GBRBM
from load_data import load_data
from load_data_2 import load_data_2
from nnClass import NN
theano.config.floatX = 'float32'


# start-snippet-1
class DBN(object):
    """Deep Belief Network

    A deep belief network is obtained by stacking several RBMs on top of each
    other. The hidden layer of the RBM at layer `i` becomes the input of the
    RBM at layer `i+1`. The first layer RBM gets as input the input of the
    network, and the hidden layer of the last RBM represents the output. When
    used for regression, the DBN is treated as a MLP, by adding a linear
    regression layer on top.
    """

    def __init__(self, numpy_rng, theano_rng=None, n_ins=645,
                 hidden_layers_sizes=[500, 500], n_outs=129,dropout_rates=[0.1,0.2,0.2],
                dropout_dist = 'binomial'):
        """This class is made to support a variable number of layers.

        :type numpy_rng: numpy.random.RandomState
        :param numpy_rng: numpy random number generator used to draw initial
                    weights

        :type theano_rng: theano.tensor.shared_randomstreams.RandomStreams
        :param theano_rng: Theano random generator; if None is given one is
                           generated based on a seed drawn from `rng`

        :type n_ins: int
        :param n_ins: dimension of the input to the DBN

        :type hidden_layers_sizes: list of ints
        :param hidden_layers_sizes: intermediate layers size, must contain
                               at least one value

        :type n_outs: int
        :param n_outs: dimension of the output of the network
        """

        self.sigmoid_layers = []
        self.rbm_layers = []
        self.dropout_layers = []
        
        self.params = []
        self.dropout_params = []
        self.n_layers = len(hidden_layers_sizes)
        
        self.param_updates = []

        assert self.n_layers > 0

        if not theano_rng:
            theano_rng = MRG_RandomStreams(numpy_rng.randint(2 ** 30))
        
        # Do not scale weights by p if "dropout distribution"
        # is Gaussian.
        w_scale = []
        for p in dropout_rates:
            if dropout_dist == 'normal':
                w_scale.append(p*0.)
            else:
                w_scale.append(p)

        # allocate symbolic variables for the data
        self.x = T.matrix('x')  # matrix with input data
        self.y = T.matrix('y')  # matrix with target vectors
                                
        # end-snippet-1
        
        
        # list with L2 regularization terms from hidden layers
        self.L2_list = []
        
        for i in xrange(self.n_layers):
            # construct the sigmoidal layer

            # the size of the input is either the number of hidden
            # units of the layer below or the input size if we are on
            # the first layer. For dropout the input elements are multiplied
            # with a stochastic variable (usually Bernoulli)
            if i == 0:
                input_size = n_ins
                layer_input = self.x
                next_dropout_layer_input = _dropout_from_layer(numpy_rng, 
                                                               layer_input, 
                                                               p=dropout_rates[0],
                                                               dist = dropout_dist)
            else:
                input_size = hidden_layers_sizes[i - 1]
                layer_input = self.sigmoid_layers[-1].output
                next_dropout_layer_input = self.dropout_layers[-1].output
            
            # DropoutHidden is the same as HiddenLayer, only with
            # output units multiplied by stochastic variables
            next_dropout_layer = DropoutHiddenLayer(rng=numpy_rng,
                    input=next_dropout_layer_input,
                    n_in=input_size, n_out=hidden_layers_sizes[i], 
                    dropout_rate=dropout_rates[i+1],
                    activation=T.nnet.sigmoid,
                    dist = dropout_dist)
            
            # The fully connected hidden layer has parameters
            # equal to the dropout layer, except weights are scaled by 
            # the expected value of the dropout stochastic variable
            # meaning (1-dropout_rate) for bernoulli, and 1 for 
            # Gaussian
            sigmoid_layer = HiddenLayer(rng=numpy_rng,
                                        input=layer_input,
                                        n_in=input_size,
                                        n_out=hidden_layers_sizes[i],
                                        W=next_dropout_layer.W * (1 - w_scale[i]),
                                        b=next_dropout_layer.b,
                                        activation=T.nnet.sigmoid)
            
            
            
            self.dropout_layers.append(next_dropout_layer)
            self.dropout_params.extend(next_dropout_layer.params)
            
            

            # add the layer to our list of layers
            self.sigmoid_layers.append(sigmoid_layer)
            
            # add the L2 regularization term for the weights of the current layer
            self.L2_list.append(sigmoid_layer.L2)

            # its arguably a philosophical question...  but we are
            # going to only declare that the parameters of the
            # sigmoid_layers are parameters of the DBN. The visible
            # biases in the RBM are parameters of those RBMs, but not
            # of the DBN.
            self.params.extend(sigmoid_layer.params)
            
            if i == 0:
                rbm_layer = GBRBM(numpy_rng=numpy_rng,
                            theano_rng=theano_rng,
                            input=layer_input,
                            n_visible=input_size,
                            n_hidden=hidden_layers_sizes[i],
                            W=next_dropout_layer.W, 
                            hbias=next_dropout_layer.b)
            else:
                rbm_layer = RBM(numpy_rng=numpy_rng,
                            theano_rng=theano_rng,
                            input=layer_input,
                            n_visible=input_size,
                            n_hidden=hidden_layers_sizes[i],
                            W=next_dropout_layer.W, 
                            hbias=next_dropout_layer.b)
            
            self.rbm_layers.append(rbm_layer)
        
        # Final layer is a linear regression layer
        dropout_output_layer = LinearRegression(
            rng =numpy_rng,
            input=self.dropout_layers[-1].output,
            n_in=hidden_layers_sizes[-1], 
            n_out=n_outs)
        
        self.dropout_layers.append(dropout_output_layer)
        self.dropout_params.extend(dropout_output_layer.params)
        
        # Final layer for fully connected DNN 
        self.linLayer = LinearRegression(
            rng = numpy_rng,
            input=self.sigmoid_layers[-1].output,
            n_in=hidden_layers_sizes[-1],
            n_out=n_outs,
            W = dropout_output_layer.W*(1 - w_scale[-1]),
            b = dropout_output_layer.b)
        
        self.params.extend(self.linLayer.params)
        
        # make list of velocity terms for use in momentum
        for param in self.dropout_params:
            velocity = theano.shared(numpy.zeros(param.get_value().shape,
                                                 dtype = theano.config.floatX)
                                    )
            self.param_updates.append(velocity)
            
        # define L2 regularization term to be sum of all layers L2 terms
        # meaning the sum of squared weights for that layer
        self.L2_reg = self.linLayer.L2 + sum(self.L2_list)

        # compute the cost for second phase of training, defined as the MSE 
        self.finetune_cost = self.linLayer.MSE(self.y)
        
        self.dropout_cost = self.dropout_layers[-1].MSE(self.y)

    def GV_est(self,train_set_x,train_set_y):
        estimate = self.linLayer.y_pred
        
        GVref = T.mean((self.y-T.mean(self.y,axis=0))**2,axis=0)
        GVref_const = T.mean((self.y-T.mean(self.y))**2)
        GVest = T.mean((estimate-T.mean(estimate,axis=0))**2,axis=0)
        GVest_const = T.mean((estimate-T.mean(estimate))**2)
        
        calc_GV = theano.function([],outputs=[GVref, GVref_const,GVest, GVest_const],givens= {self.x: train_set_x,
                                                                    self.y: train_set_y})
        return calc_GV()
    
#    # TODO: write code for max norm regularization        
    def max_norm_update(self,new_param,c, epsilon=1e-7):
        # find lenght of incoming weight vectors
        norms = T.sqrt(T.sum(new_param**2,axis = 0, keepdims=True))
        
        dtype = numpy.dtype(theano.config.floatX).type

        #restrict norms to be no greater than c
        target_norms = T.clip(norms, 0, dtype(c))
        
        scaled_param = new_param*(target_norms / (norms + dtype(epsilon))) 
        return scaled_param
            
    

    
    def pretraining_functions(self, train_set_x, batch_size, k):
        '''Generates a list of functions, for performing one step of
        gradient descent at a given layer. The function will require
        as input the minibatch index, and to train an RBM you just
        need to iterate, calling the corresponding function on all
        minibatch indexes.

        :type train_set_x: theano.tensor.TensorType
        :param train_set_x: Shared var. that contains all datapoints used
                            for training the RBM
        :type batch_size: int
        :param batch_size: size of a [mini]batch
        :param k: number of Gibbs steps to do in CD-k / PCD-k

        '''

        # index to a [mini]batch
        index = T.lscalar('index')  # index to a minibatch
        learning_rate = T.scalar('lr')  # learning rate to use
        
        # number of batches
        n_batches = train_set_x.get_value(borrow=True).shape[0] / batch_size

        # begining of a batch, given `index`
        batch_begin = index * batch_size
        # ending of a batch given `index`
        batch_end = batch_begin + batch_size

        pretrain_fns = []
        rbm_numb=0
        for rbm in self.rbm_layers:

            # get the cost and the updates list
            # using CD-k here (persisent=None) for training each RBM.
            # TODO: change cost function to reconstruction error
            cost, updates = rbm.get_cost_updates(learning_rate,
                                                 persistent=None, k=k)
            

            # compile the theano function
            #theano.In(learning_rate, value=0.1)
            fn = theano.function(
                inputs=[index, learning_rate],
                outputs=cost,
                updates=updates,
                givens={
                    self.x: train_set_x[batch_begin:batch_end]
                }
            )
            # append `fn` to the list of functions
            pretrain_fns.append(fn)

        return pretrain_fns
    
    def monitor_overfitting(self, train_set_x, valid_set_x):
        """
        Returns two functions for every RBM, that calculates 
        the free energy (FE) of the RBM using training and 
        validation data. The number of data points used
        is equal to half the size of the validation set for
        both the training data and the validation data.
        
        The purpose of returning the free energies is to 
        monitor overfitting. The difference:
        
        FE_valid - FE_train 
        
        is equal to the natural logarithm of the probability ratio:
        
        P(training data; theta_RBM) / P(validation data; theta_RBM)
        
        where theta_RBM are the RBM parameters. 
        If this value is rising (after some time) it could 
        indicate that the model is overfitting.
        
        """
        train_data = T.matrix('train_data')
        valid_data = T.matrix('valid_data')
        
        function_list =[]
        
        for rbm in self.rbm_layers:
            
            fe = T.mean(rbm.free_energy(rbm.input))
            
            
            FE_train = theano.function(
                inputs=[],
                outputs=fe,
                givens={
                    self.x: train_set_x
                }
            )
            
            FE_valid = theano.function(
                inputs=[],
                outputs=fe,
                givens={
                    self.x: valid_set_x
                }
            )
            function_list.extend([FE_train,FE_valid])
         
        return function_list
        
        
    def build_finetune_functions(self, datasets, batch_size=128, W_decay=0.00001,mom_method='tutorial',c=3,MNreg=False):
        '''Generates a function `train` that implements one step of
        finetuning, a function `validate` that computes the error on a
        batch from the validation set, and a function `test` that
        computes the error on a batch from the testing set

        :type datasets: list of pairs of theano.tensor.TensorType
        :param datasets: It is a list that contain all the datasets;
                        the has to contain three pairs, `train`,
                        `valid`, `test` in this order, where each pair
                        is formed of two Theano variables, one for the
                        datapoints, the other for the labels
        :type batch_size: int
        :param batch_size: size of a minibatch
        :type learning_rate: float
        :param learning_rate: learning rate used during finetune stage

        '''

        (train_set_x, train_set_y) = datasets[0]
        (valid_set_x, valid_set_y) = datasets[1]
        (test_set_x, test_set_y) = datasets[2]

        # compute number of minibatches for training, validation and testing
        n_train_batches = train_set_x.get_value(borrow=True).shape[0]
        n_train_batches /= batch_size
        n_valid_batches = valid_set_x.get_value(borrow=True).shape[0]
        n_valid_batches /= batch_size
        n_test_batches = test_set_x.get_value(borrow=True).shape[0]
        n_test_batches /= batch_size

        index = T.lscalar('index')  # index to a [mini]batch
        momentum = T.scalar('momentum')
        learning_rate = T.scalar('learning_rate')

        
        # compute the gradients with respect to the model parameters inc reg.
        #gparams = T.grad(self.finetune_cost+W_decay/2*self.L2_reg, self.params)
        
        gparams = T.grad(self.dropout_cost, self.dropout_params)
        # compute list of fine-tuning updates including momentum (weird implementation)
        updates= []
        
        if mom_method == 'classic':
            parcount = 0
            for param, prev_update, grad in zip(self.dropout_params,self.param_updates,gparams):
                #find updateterm
                updateterm = momentum*prev_update-learning_rate*grad
                
                # for weight matrices use max norm regularization
                if parcount%2 == 0 and MNreg:
                    new_param = self.max_norm_update(param+updateterm,c)
                else:
                    new_param = param+updateterm
           
                # add update expression for parameters
                updates.append((param,new_param))
            
                # add update expression for previous update/velocity:
                updates.append((prev_update, updateterm))
                
                parcount += 1
                
        elif mom_method == 'tutorial':
            parcount = 0
            for param, prev_update, grad in zip(self.dropout_params,self.param_updates,gparams):
                #find updateterm
                updateterm = (momentum*prev_update+(1.-momentum)*grad)
                
                # for weight matrices use max norm regularization
                if parcount%2 == 0 and MNreg:
                    new_param = self.max_norm_update(param-learning_rate*updateterm,c)
                else:
                    new_param = param-(learning_rate+0.5)*updateterm
           
                # add update expression for parameters
                updates.append((param,new_param))
            
                # add update expression for previous update/velocity:
                updates.append((prev_update, updateterm))
                
                parcount += 1
                
            
        else: # 'weird'
            parcount = 0
            for param, prev_update, grad in zip(self.dropout_params,self.param_updates,gparams):
                #find updateterm
                updateterm = learning_rate*(momentum*prev_update-(1.-momentum)*grad)
                
                # for weight matrices use max norm regularization
                if parcount%2 == 0 and MNreg:
                    new_param = self.max_norm_update(param+updateterm,c)
                else:
                    new_param = param+updateterm
           
                # add update expression for parameters
                updates.append((param,new_param))
            
                # add update expression for previous update/velocity:
                updates.append((prev_update, updateterm))
                
                parcount += 1
    
        
#        
#        for i in range(len(self.params)):
#           #self.param_updates[i] = momentum*self.param_updates[i]-learning_rate*gparams[i]

#            # update parameters
#            updates.append((self.params[i],self.params[i]-learning_rate*gparams[i]+momentum*self.param_updates[i]))
#            
#            updates.append((self.param_updates[i],momentum*self.param_updates[i]-learning_rate*gparams[i]))
        
                           
        
        print(updates)                  

        train_fn = theano.function(
            inputs=[index,momentum,learning_rate],
            outputs=[self.dropout_cost,self.finetune_cost],
            updates=updates,
            givens={
                self.x: train_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: train_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )

        test_score_i = theano.function(
            [index],
            outputs=self.linLayer.MSE(self.y),
            givens={
                self.x: test_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: test_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )

        valid_score_i = theano.function(
            [index],
            outputs=self.linLayer.MSE(self.y),
            givens={
                self.x: valid_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: valid_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )
        
        train_score_i = theano.function(
            [index],
            outputs=self.linLayer.MSE(self.y),
            givens={
                self.x: train_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: train_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )
        
        # Create a function that scans the entire validation set
        def valid_score():
            return [valid_score_i(i) for i in xrange(n_valid_batches)]

        # Create a function that scans the entire test set
        def test_score():
            return [test_score_i(i) for i in xrange(n_test_batches)]
        
        # Create a function that scans the entire training set
        def train_score():
            return [train_score_i(i) for i in xrange(n_train_batches)]

        return train_fn, valid_score, test_score, train_score


def _dropout_from_layer(rng, layer, p,dist):
    """p is the probablity of dropping a unit
    """
    srng = theano.tensor.shared_randomstreams.RandomStreams(
            rng.randint(999999))
    # p=1-p because 1's indicate keep and p is prob of dropping
    if dist =='normal':
        mask = srng.normal(size=layer.shape,avg=1.0,std = numpy.sqrt(p/(1-p)))
    else:
        mask = srng.binomial(n=1, p=1-p, size=layer.shape)
    # The cast is important because
    # int * float32 = float64 which pulls things off the gpu
    output = layer * T.cast(mask, theano.config.floatX)
    return output

class DropoutHiddenLayer(HiddenLayer):
    def __init__(self, rng, input, n_in, n_out,
                 activation, dropout_rate,dist = 'binomial', W=None, b=None):
        super(DropoutHiddenLayer, self).__init__(
                rng=rng, input=input, n_in=n_in, n_out=n_out, W=W, b=b,
                activation=activation)

        self.output = _dropout_from_layer(rng, self.output, p=dropout_rate,dist = dist)
        

def train_DBN(finetune_lr=0.1, pretraining_epochs=50,
             pretrain_lr=0.0005, k=1, training_epochs=200,
             batch_size=128,finetune_wd = 0.,
             momentum = 0.5,momentum_method = 'tutorial',
             use_MN = True,max_norm=4.,use_pretrained_params=False):
    """
    Demonstrates how to train and test a Deep Belief Network.

    :type finetune_lr: float
    :param finetune_lr: learning rate used in the finetune stage
    :type pretraining_epochs: int
    :param pretraining_epochs: number of epoch to do pretraining
    :type pretrain_lr: float
    :param pretrain_lr: learning rate to be used during pre-training
    :type k: int
    :param k: number of Gibbs steps in CD/PCD
    :type training_epochs: int
    :param training_epochs: maximal number of iterations ot run the optimizer
    :type dataset: string
    :param dataset: path the the pickled dataset
    :type batch_size: int
    :param batch_size: the size of a minibatch
    :type finetune_wd: float
    :param finetune_wd: weight decay (L2_reg) coefficient
    :type momentum: float
    :param momentum: 0<= momentum <1 
    """
    writedict ={}
    Nfrq = 129
    left_cfr = 2
    right_cfr = 2
    numb_cfr = left_cfr + 1 + right_cfr 
    n_ins = numb_cfr*129
    SNRlist = [0,5,10]
    
    dropout_dist = 'bernoulli'
    
    
    numb_train_fr = 0.5*224896 # 224896 frames is about 1 hour
    numb_valid_fr = 0.5*74880
    numb_test_fr = 0.5*74880
    
    noisetypes = ['AWGN','babble','restaurant','street']
    validnoise = ['AWGN','babble','restaurant','street']
    testnoise = ['car','exhibition']
    
    #loading data:
    print('loading data...')
    train_load_x, train_load_y,next_line_tr = load_data_2(filelist = 'train_stem_big.list',
                                             path = 'training_preprosd/',
                                             Nframes = numb_train_fr,
                                             left_cfr = left_cfr,
                                             right_cfr = right_cfr,
                                             Nfrq = Nfrq,
                                             SNRlist = SNRlist,
                                             noisetypes = noisetypes,
                                             shuffle_whole = True,
                                             start_line = 0)
    
    print('read %i lines from TIMIT for training set' % (next_line_tr-1))
                                         
    valid_load_x, valid_load_y,next_line_val = load_data_2(filelist = 'test_overlap_stem_big.list',
                                             path = 'validation_preprosd/',
                                             Nframes = numb_valid_fr,
                                             left_cfr = left_cfr,
                                             right_cfr = right_cfr,
                                             Nfrq = Nfrq,
                                             SNRlist = SNRlist,
                                             noisetypes = validnoise,
                                             shuffle_whole = True,
                                             start_line = 0)
    
    print('read %i lines from TIMIT for validation set' % (next_line_val-1))
                                         
    test_load_x, test_load_y,next_line_test = load_data_2(filelist = 'test_separate_stem.list',
                                           path = 'test_preprosd/',
                                           Nframes = numb_test_fr,
                                           left_cfr = left_cfr,
                                           right_cfr = right_cfr,
                                           Nfrq = Nfrq,
                                           SNRlist = SNRlist,
                                           noisetypes = testnoise,
                                           shuffle_whole = True,
                                           start_line = 0)
    
    print('read %i lines from TIMIT for test set' % (next_line_test-1))
    
    traindat_sec = train_load_x.shape[0]*(16/1000.)
    traindat_min = traindat_sec/60.
    traindat_hou = int(traindat_min/60.)
    traindat_min_rest = int(traindat_min-traindat_hou*60.)
    traindat_sec_rest = int(traindat_sec-int(traindat_min)*60.)
   
    
    print('training data is %i number of frames, equaling %d mini-batches' 
          % (train_load_x.shape[0],train_load_x.shape[0]/128.))
    print('This euqals about %i hours, %i minutes and %i seconds of speech data' 
          % (traindat_hou,traindat_min_rest,traindat_sec_rest))
    
    #normalize data (keep normalization factors)
    
    datamean_in = numpy.mean(train_load_x, axis = 0)
    datastd_in = numpy.std(train_load_x, axis = 0,ddof=1)
    
    datamean_out = datamean_in[(numb_cfr-1)/2*Nfrq:((numb_cfr-1)/2+1)*Nfrq]
    datastd_out = datastd_in[(numb_cfr-1)/2*Nfrq:((numb_cfr-1)/2+1)*Nfrq]
    
    if not (datastd_in==0).sum():
        print('scaling data...')
        for i in range(numb_cfr*Nfrq):
            train_load_x[:][:,i] = (train_load_x[:][:,i]-datamean_in[i]) / datastd_in[i]
            valid_load_x[:][:,i] = (valid_load_x[:][:,i]-datamean_in[i]) / datastd_in[i]
            test_load_x[:][:,i] = (test_load_x[:][:,i]-datamean_in[i]) / datastd_in[i]
                
        for i in range(Nfrq):
            train_load_y[:][:,i] = (train_load_y[:][:,i]-datamean_out[i]) / datastd_out[i]
            valid_load_y[:][:,i] = (valid_load_y[:][:,i]-datamean_out[i]) / datastd_out[i]
            test_load_y[:][:,i] = (test_load_y[:][:,i]-datamean_out[i]) / datastd_out[i]
    else:
        sys.exit('Error! Zero in feature scaling factor')
    

    
    #store data in theano.shared variables
    print('storing data in theano.shared variables...')
    train_set_x = theano.shared(train_load_x,borrow=True)
    valid_set_x = theano.shared(valid_load_x,borrow=True)
    test_set_x = theano.shared(test_load_x,borrow=True)
                                       
    train_set_y = theano.shared(train_load_y,borrow=True)
    valid_set_y = theano.shared(valid_load_y,borrow=True)
    test_set_y = theano.shared(test_load_y,borrow=True)
    
    #define list of datasets as used in example code
    datasets = [(train_set_x, train_set_y), (valid_set_x, valid_set_y),
                (test_set_x, test_set_y)]
                
    # compute number of minibatches for training, validation and testing
    n_train_batches = train_set_x.get_value(borrow=True).shape[0] / batch_size
    


    # numpy random generator
    numpy_rng = numpy.random.RandomState(5098)
    print('... building the model')
    
    # construct the Deep Belief Network
 
    hidden_sizes = [875,875,875]
    
    dropout_rates=[0.1,0.2,0.2,0.2] 
    
    w_scale = []
    for p in dropout_rates:
        if dropout_dist == 'normal':
            w_scale.append(p*0.)
        else:
            w_scale.append(p)
    
    dbn = DBN(numpy_rng=numpy_rng, n_ins=n_ins,
              hidden_layers_sizes=hidden_sizes,
              n_outs=129,dropout_rates=dropout_rates,
              dropout_dist = dropout_dist)
    
    # for storing best NN given validation set performance
    dbn_best = DBN(numpy_rng=numpy_rng, n_ins=n_ins,
              hidden_layers_sizes=hidden_sizes,
              n_outs=129,dropout_rates=dropout_rates,
              dropout_dist = dropout_dist)

    # for storing best NN given training set performance
    # to be used for resetting momentum if "rolling past"
    # minimum when using momentum
    dbn_best_training = DBN(numpy_rng=numpy_rng, n_ins=n_ins,
              hidden_layers_sizes=hidden_sizes,
              n_outs=129,dropout_rates=dropout_rates,
              dropout_dist = dropout_dist)
    
    
    
    # start-snippet-2
    #########################
    # PRETRAINING THE MODEL #
    #########################
    
    
    print('... getting the pretraining functions')
    pretraining_fns = dbn.pretraining_functions(train_set_x=train_set_x,
                                                batch_size=batch_size,
                                                k=k)
    monitoring_fn = dbn.monitor_overfitting(
        train_set_x=train_set_x[0:150],
        valid_set_x=valid_set_x[0:150]
    )
    
    if use_pretrained_params:
        pretraining_epochs=0
          
    
    PTrec_cost = {}
    FEdiffs = {}
    print('... pre-training the model')
    start_time = timeit.default_timer()
    ## Pre-train layer-wise
    for i in xrange(dbn.n_layers):
        # go through pretraining epochs
        LRscale = 1.
        PT_epoch_scale = 1.
        
        
        if i==0:
            rec_cost = '(MSE)'
            LRscale = LRscale/10.
        else:
            rec_cost = '(cross-entropy)'
            #try smaller amount of epochs for Bernoulli-Bernoulli RBMs
            PT_epoch_scale = 1. 
        PTrec_cost['PT_layer_' + str(i)] = []
        FEdiffs['FEdiff_layer_' + str(i)] = []
        
        for epoch in xrange(int(PT_epoch_scale*pretraining_epochs)):
            # go through the training set
            c = []
            #W_e = sum(sum(dbn.dropout_params[2*i].get_value()**2))
            #print(W_e)
            for batch_index in xrange(n_train_batches):
                c.append(pretraining_fns[i](index=batch_index,
                                            lr=LRscale*pretrain_lr))
            print('Pre-training layer %i, epoch %d, cost %s ' % (i, epoch,rec_cost)),
            print(numpy.mean(c))
            PTrec_cost['PT_layer_' + str(i)].append(numpy.mean(c))
            
            #if reconstruction cost is increased by 10. reduce learning rate by 50% and use MNreg
            if epoch>2:
                if PTrec_cost['PT_layer_' + str(i)][-1]/PTrec_cost['PT_layer_' + str(i)][-2] > 5.:
                    LRscale = LRscale/2. 
                    print('Rescaled learning rate to ' + str(LRscale))
            
            #calculate difference in free energy for training set and validation set
            f1 = monitoring_fn[2*i]() #free energy of training set
            f2 = monitoring_fn[2*i+1]() #free energy of validation set
            diff = f2-f1
            FEdiffs['FEdiff_layer_' + str(i)].append(diff)
            
            #print('Free energy difference between training and validation is %s ' % diff)
            
    end_time = timeit.default_timer()
    
    
    if use_pretrained_params:
        print('...loading pretrained parameters')
        f = open('parameters/pretrained_params.save','rb')
        dbn = cPickle.load(f)
        f.close()
    else: #store pretrained params
        print('Saving pretrained parameters')
        f = open('parameters/pretrained_params.save','wb')
        cPickle.dump(dbn, f, protocol=cPickle.HIGHEST_PROTOCOL)
        f.close()       

                
            
        
    
    # end-snippet-2
    print >> sys.stderr, ('The pretraining code for file ' +
                          os.path.split(__file__)[1] +
                          ' ran for %.2fm' % ((end_time - start_time) / 60.))

    pt_params={}
    pt_params['W0'] = dbn.dropout_params[0].get_value(borrow=True)*(1-w_scale[0])
    scipy.io.savemat('parameters/PT_DBN_layer1.mat',pt_params)
    
    ########################
    # FINETUNING THE MODEL #
    ########################
    
    #scaling results from pretraining
    if dropout_dist != 'normal':
        print('Scaling pretrained parameters')
        num = 0
        for param in dbn.dropout_params:
            if num%2==0:
                param.set_value(param.get_value()/(1-w_scale[int(num/2)]))
            num +=1
        
    

    # get the training, validation and testing function for the model
    print('... getting the finetuning functions')
    train_fn, validate_model, test_model, train_performance = dbn.build_finetune_functions(
        datasets=datasets,
        batch_size=batch_size,
        W_decay = finetune_wd,
        mom_method = momentum_method,
        c=max_norm,
        MNreg=use_MN
    )

    print('... finetuning the model')

    improvement_threshold = 0.995  # a relative improvement of this much is
                                   # considered significant
    validation_frequency = n_train_batches
                                  # go through this many
                                  # minibatches before checking the network
                                  # on the validation set; in this case we
                                  # check every epoch

    best_validation_loss = numpy.inf
    best_training_loss = numpy.inf
    best_params_testscore = numpy.inf
    test_score = 0.
    start_time = timeit.default_timer()

    done_looping = False
    epoch = 0
    best_iter = 0
    
    #minibatch_cost_dev = []
    training_cost_dev = []
    validation_cost_dev = []
    test_cost_dev = []
    
    momentum_dev = []
    maxmom = 0.9
    mom_inc = True
    
    FT_lr =finetune_lr
    lr_scale = 0.95
    dtype = numpy.dtype(theano.config.floatX).type
    
    #final_layer_norms = []
    
    while (epoch < training_epochs) and (not done_looping):
        
        
        if epoch == 0:
            training_losses = train_performance()
            this_training_loss = numpy.mean(training_losses)
            training_cost_dev.append(this_training_loss)

            test_losses = test_model()
            test_score = numpy.mean(test_losses)
            test_cost_dev.append(test_score)

            validation_losses = validate_model()
            this_validation_loss = numpy.mean(validation_losses)
            validation_cost_dev.append(this_validation_loss)
            print(' Initial training loss %s, validation loss %s, and test loss %s' 
                  % (
                    this_training_loss,
                    this_validation_loss,
                    test_score)
                  )
            
        epoch = epoch + 1
        
        # supervise norms of incoming weight vectors to firs layer
        #E_w = numpy.transpose(numpy.sqrt(numpy.sum(dbn.params[0].get_value()**2,axis=0,keepdims=True)))
        #final_layer_norms.append(E_w)
        
        for minibatch_index in xrange(n_train_batches):
            
            minibatch_avg_cost = train_fn(minibatch_index,momentum,FT_lr)
            #minibatch_cost_dev.append(minibatch_avg_cost)
            iter = (epoch - 1) * n_train_batches + minibatch_index

#            momentumval = numpy.minimum(1-1/(2*(numpy.floor(iter/250)+1)),maxmom)
#            momentum = numpy.asarray(momentumval,dtype = theano.config.floatX)
            if (iter + 1) % validation_frequency == 0:

                validation_losses = validate_model()
                this_validation_loss = numpy.mean(validation_losses)
                validation_cost_dev.append(this_validation_loss)
                print(
                    'epoch %i, minibatch %i/%i, validation cost  %f'
                    % (
                        epoch,
                        minibatch_index + 1,
                        n_train_batches,
                        this_validation_loss
                    )
                )
                
#                if minibatch_index==0:
#                    done_looping = True
#                    break
                # test it on the test set
                test_losses = test_model()
                test_score = numpy.mean(test_losses)
                test_cost_dev.append(test_score)
            
                # if we got the best validation score until now
                if this_validation_loss < best_validation_loss*improvement_threshold:

                    # save best validation score and iteration number
                    best_validation_loss = this_validation_loss
                    best_iter = iter
                    best_params_testscore = test_score
                    
                    # store parameter values in dbn_best
                    for paramnumb in range(len(dbn.params)):
                        dbn_best.dropout_params[paramnumb].set_value(dbn.dropout_params[paramnumb].get_value(borrow=True))
                    
                        
                    print(('     epoch %i, minibatch %i/%i, test cost of '
                           'best model %f') %
                          (epoch, minibatch_index + 1, n_train_batches,
                           test_score))
             
        
        #increase momentum
        if mom_inc and epoch>10 and momentum<=maxmom:
            momentum += 0.005
            if momentum<0.5:
                momentum = 0.5
                
        if momentum>=maxmom: #after momentum is max, start scaling learning rate
            FT_lr = dtype(FT_lr*lr_scale)

        # find full-batch training set MSE
        training_losses = train_performance()
        this_training_loss = numpy.mean(training_losses)
        training_cost_dev.append(this_training_loss)
                
        #store best training loss
        if this_training_loss < best_training_loss*improvement_threshold:
            best_training_loss = this_training_loss
            for paramnumb in range(len(dbn.params)):
                dbn_best_training.dropout_params[paramnumb].set_value(dbn.dropout_params[paramnumb].get_value(borrow=True))
        else: 
            # if training cost is 5 times the lowest so far, reset to best parameters
            # and set momentum to 0
            if this_training_loss/best_training_loss > 5.:
                # set momentum to clear history in velocity term 
                momentum = 0.
                        
                # updating parameters to equal best performance on training set so far
                for paramnumb in range(len(dbn.params)):
                    dbn.dropout_params[paramnumb].set_value(dbn_best_training.dropout_params[paramnumb].get_value(borrow=True))
                
                
        momentum_dev.append(momentum)
        
    end_time = timeit.default_timer()
    
    
    
    print(
        (
            'Optimization complete with best validation score of %f %%, '
            'obtained at iteration %i, '
            'with test performance %f'
        ) % (best_validation_loss, best_iter + 1, best_params_testscore)
    )
    print >> sys.stderr, ('The fine tuning code for file ' +
                          os.path.split(__file__)[1] +
                          ' ran for %.2fm' % ((end_time - start_time)
                                              / 60.))
    
    test_set_x.set_value([[]])
    test_set_y.set_value([[]])
    valid_set_x.set_value([[]])
    valid_set_y.set_value([[]])
    
    GVref, GVref_const , GVest , GVest_const= dbn_best.GV_est(train_set_x[0:int(train_set_x.get_value().shape[0]/2)],
                                                              train_set_y[0:int(train_set_x.get_value().shape[0]/2)])
    GVeq = numpy.sqrt(GVref/GVest)
    GVeq_const = numpy.sqrt(GVref_const/GVest_const)
    
    #avoid scaling weights if gaussian dropout is used
    
    
    matlabs = {} #for saving dbn.params for matlab
    matlabs2 = {} #for saving dbn_best.params for matlab
    
    # save parameters
    #save_params = open('parameters/dbn.save','wb')
    #cPickle.dump(dbn,save_params,-1)
    #cPickle.dump(datamean_in,save_params,-1)
    #cPickle.dump(datastd_in,save_params,-1)
    paramtype = ['W','b']
    
    matlabs['my_in'] = datamean_in
    matlabs['sigma_in'] = datastd_in
    matlabs['my_tar'] = datamean_out
    matlabs['sigma_tar'] = datastd_out
    
    matlabs2['my_in'] = datamean_in
    matlabs2['sigma_in'] = datastd_in
    matlabs2['my_tar'] = datamean_out
    matlabs2['sigma_tar'] = datastd_out
    #matlabs2['train_mb_dev'] = minibatch_cost_dev
    matlabs2['training_dev'] = training_cost_dev
    matlabs2['validation_dev'] = validation_cost_dev
    matlabs2['test_dev'] = test_cost_dev
    matlabs2['momentum_dev'] = momentum_dev
    matlabs2['GVref'] = GVref
    matlabs2['GVest'] = GVest
    matlabs2['GVeq'] = GVeq
    matlabs2['GVref_const'] = GVref_const
    matlabs2['GVest_const'] = GVest_const
    matlabs2['GVeq_const'] = GVeq_const
    #matlabs2['norm_dev'] = final_layer_norms
    for i in range(len(dbn.params)):
        if i%2==0: #parameter is weight matrix
            matlabs[paramtype[i%2]+str(int(i/2))]=dbn.dropout_params[i].get_value(borrow=True)*(1 - w_scale[int(i/2)])
            matlabs2[paramtype[i%2]+str(int(i/2))]=dbn_best.dropout_params[i].get_value(borrow=True)*(1 - w_scale[int(i/2)])
        else:
            matlabs[paramtype[i%2]+str(int(i/2))]=dbn.dropout_params[i].get_value(borrow=True)
            matlabs2[paramtype[i%2]+str(int(i/2))]=dbn_best.dropout_params[i].get_value(borrow=True)
    
    
    #save pretrain-costs
    for key1,key2 in zip(PTrec_cost, FEdiffs):
        matlabs2[key1] = PTrec_cost[key1]
        matlabs2[key2] = FEdiffs[key2]
        
    
    
    
    #save_params.close()
    scipy.io.savemat('parameters/formatlab.mat',matlabs)
    scipy.io.savemat('parameters/formatlab_best.mat',matlabs2)
    
    # storing metadata
    metadata = []
    
    metadata.append('PT_lr = ' + str(pretrain_lr) + '\n')
    metadata.append('PT_epochs = ' + str(pretraining_epochs) + '\n')
    metadata.append('FT_lr = ' + str(finetune_lr) + '\n')
    metadata.append('FT_epochs = ' + str(training_epochs) + '\n')
    metadata.append('momentum_method = ' + str(momentum_method) + '\n')
    metadata.append('use_MN = ' + str(use_MN) + ' with max_norm ' + str(max_norm)+'\n')
    
    
    metadata.append('number of training frames = ' + str(numb_train_fr) + ' = about ' +str(numb_train_fr/224896) +  ' hours\n')
    metadata.append('number of context frames in input, left = ' + str(left_cfr) + ' right = ' + str(right_cfr) + '\n')
    metadata.append('number of TIMIT files used for training = ' + str(next_line_tr) + '\n')
    metadata.append('number of TIMIT files used for validation = ' + str(next_line_val) + '\n')
    metadata.append('number of TIMIT files used for test = ' + str(next_line_test) + '\n')
    metadata.append('noise types in training set = ' + str(noisetypes) +'\n')
    metadata.append('noise types in validation set = ' + str(validnoise) +'\n')
    metadata.append('test noise types= ' + str(testnoise) +'\n')
    metadata.append('SNR levels= ' + str(SNRlist) +'\n')
    metadata.append('hidden layer sizes = ' + str(hidden_sizes) +'\n')
    metadata.append('dropout rates = ' + str(dropout_rates) +'\n')
    metadata.append('dropout distibution = ' + str(dropout_dist) +'\n')
    metadata.append('learning rate scaling= ' + str(lr_scale) +'\n')
    
    metadatafile = open('parameters/DBN_metadata.txt','w')
    for metaline in metadata:
        metadatafile.write(metaline)
        
    metadatafile.close()



if __name__ == '__main__':
    train_DBN()