# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 11:54:13 2016

@author: deeplearning.net, edited for regression by OKM
"""

"""
"""
import os
import sys
import timeit
import cPickle

import numpy
import scipy.io

import theano
import theano.tensor as T
from theano.sandbox.rng_mrg import MRG_RandomStreams

from linregClass import LinearRegression
from MLPclass import HiddenLayer
from load_data import load_data
from load_data_2 import load_data_2
from nnClass import NN
theano.config.floatX = 'float32'


# start-snippet-1
class DNN(object):

    def __init__(self, numpy_rng, theano_rng=None, n_ins=645,
                 hidden_layers_sizes=[500, 500], n_outs=129,
                 dropout_rates=[0.1,0.2,0.2],normal_init=False,
                 dropout_dist = 'binomial'):


        self.hidden_layers = []
        self.dropout_layers = []
        
        self.params = []
        self.dropout_params = []
        self.n_layers = len(hidden_layers_sizes)
        
        

        assert self.n_layers > 0

        if not theano_rng:
            theano_rng = MRG_RandomStreams(numpy_rng.randint(2 ** 30))
        
        w_scale = []
        for p in dropout_rates:
            if dropout_dist == 'normal':
                w_scale.append(p*0.)
            else:
                w_scale.append(p)


        # allocate symbolic variables for the data
        self.x = T.matrix('x')  # matrix with input data
        self.y = T.matrix('y')  # matrix with target vectors
                                
        # end-snippet-1
        # The DBN is an MLP, for which all weights of intermediate
        # layers are shared with a different RBM.  We will first
        # construct the DBN as a deep multilayer perceptron, and when
        # constructing each hidden layer we also construct an RBM
        # that shares weights with that layer. During pretraining we
        # will train these RBMs (which will lead to chainging the
        # weights of the MLP as well) During finetuning we will finish
        # training the DBN by doing stochastic gradient descent on the
        # MLP.
        
        # list with L2 regularization terms from hidden layers
        self.L2_list = []
        
        for i in xrange(self.n_layers):
            # construct the hidden layer

            # the size of the input is either the number of hidden
            # units of the layer below or the input size if we are on
            # the first layer
            # the input to this layer is either the activation of the
            # hidden layer below or the input of the DBN if you are on
            # the first layer
            if i == 0:
                input_size = n_ins
                layer_input = self.x
                next_dropout_layer_input = _dropout_from_layer(numpy_rng, 
                                                               layer_input, 
                                                               p=dropout_rates[0],
                                                               dist =dropout_dist)
            else:
                input_size = hidden_layers_sizes[i - 1]
                layer_input = self.hidden_layers[-1].output
                next_dropout_layer_input = self.dropout_layers[-1].output
            
            next_dropout_layer = DropoutHiddenLayer(rng=numpy_rng,
                    input=next_dropout_layer_input,
                    n_in=input_size, n_out=hidden_layers_sizes[i], 
                    dropout_rate=dropout_rates[i+1],
                    activation='relu',normal_init=normal_init,
                    dist =dropout_dist)
            
            hidden_layer = HiddenLayer(rng=numpy_rng,
                                        input=layer_input,
                                        n_in=input_size,
                                        n_out=hidden_layers_sizes[i],
                                        W=next_dropout_layer.W * (1 - w_scale[i]),
                                        b=next_dropout_layer.b,
                                        activation='relu')
            
            
            
            self.dropout_layers.append(next_dropout_layer)
            self.dropout_params.extend(next_dropout_layer.params)
            
            

            # add the layer to our list of layers
            self.hidden_layers.append(hidden_layer)
            
            # add the L2 regularization term for the weights of the current layer
            self.L2_list.append(hidden_layer.L2)

            # its arguably a philosophical question...  but we are
            # going to only declare that the parameters of the
            # hidden_layers are parameters of the DBN. The visible
            # biases in the RBM are parameters of those RBMs, but not
            # of the DBN.
            self.params.extend(hidden_layer.params)
            
            # extend param_updates with appropriate number of zeros for new layer parameters
            
            
        # We now need to add a logistic layer on top of the MLP
        
        dropout_output_layer = LinearRegression(
            rng =numpy_rng,
            input=self.dropout_layers[-1].output,
            n_in=hidden_layers_sizes[-1], 
            n_out=n_outs)
        
        self.dropout_layers.append(dropout_output_layer)
        self.dropout_params.extend(dropout_output_layer.params)
        
        
        self.linLayer = LinearRegression(
            rng = numpy_rng,
            input=self.hidden_layers[-1].output,
            n_in=hidden_layers_sizes[-1],
            n_out=n_outs,
            W = dropout_output_layer.W*(1 - w_scale[-1]),
            b = dropout_output_layer.b)
        
        self.params.extend(self.linLayer.params)
        
        
        # define L2 regularization term to be sum of all layers L2 terms
        # meaning the sum of squared weights for that layer
        self.L2_reg = self.linLayer.L2 + sum(self.L2_list)
        
        

        # compute the cost for second phase of training, defined as the
        # MSE 
        self.finetune_cost = self.linLayer.MSE(self.y)
        
        self.dropout_cost = self.dropout_layers[-1].MSE(self.y)

    def GV_est(self,train_set_x,train_set_y):
        estimate = self.linLayer.y_pred
        
        GVref = T.mean((self.y-T.mean(self.y,axis=0))**2,axis=0)
        GVref_const = T.mean((self.y-T.mean(self.y))**2)
        GVest = T.mean((estimate-T.mean(estimate,axis=0))**2,axis=0)
        GVest_const = T.mean((estimate-T.mean(estimate))**2)
        
        calc_GV = theano.function([],outputs=[GVref, GVref_const,GVest, GVest_const],givens= {self.x: train_set_x,
                                                                    self.y: train_set_y})
        return calc_GV()
          
    def max_norm_update(self,new_param,c, epsilon=1e-7):
        # find lenght of incoming weight vectors
        norms = T.sqrt(T.sum(new_param**2,axis = 0, keepdims=True))
        
        dtype = numpy.dtype(theano.config.floatX).type

        #restrict norms to be no greater than c
        target_norms = T.clip(norms, 0, dtype(c))
        
        scaled_param = new_param*(target_norms / (norms + dtype(epsilon))) 
        return scaled_param
            
    
        
    def build_finetune_functions(self, datasets, batch_size=128, W_decay=0.00001,mom_method='weird',c=3,MNreg=False):
        '''Generates a function `train` that implements one step of
        finetuning, a function `validate` that computes the error on a
        batch from the validation set, and a function `test` that
        computes the error on a batch from the testing set

        :type datasets: list of pairs of theano.tensor.TensorType
        :param datasets: It is a list that contain all the datasets;
                        the has to contain three pairs, `train`,
                        `valid`, `test` in this order, where each pair
                        is formed of two Theano variables, one for the
                        datapoints, the other for the labels
        :type batch_size: int
        :param batch_size: size of a minibatch
        :type learning_rate: float
        :param learning_rate: learning rate used during finetune stage

        '''

        (train_set_x, train_set_y) = datasets[0]
        (valid_set_x, valid_set_y) = datasets[1]
        (test_set_x, test_set_y) = datasets[2]

        # compute number of minibatches for training, validation and testing
        n_train_batches = train_set_x.get_value(borrow=True).shape[0]
        n_train_batches /= batch_size
        n_valid_batches = valid_set_x.get_value(borrow=True).shape[0]
        n_valid_batches /= batch_size
        n_test_batches = test_set_x.get_value(borrow=True).shape[0]
        n_test_batches /= batch_size

        index = T.lscalar('index')  # index to a [mini]batch
        momentum = T.scalar('momentum')
        learning_rate = T.scalar('learning_rate')
        
        self.param_updates = [] 
        if mom_method =='rmsprop':
                   
            for param in self.dropout_params:
                velocity = theano.shared(numpy.ones(param.get_value().shape,
                                                     dtype = theano.config.floatX)
                                        )
                self.param_updates.append(velocity)
        else:
            for param in self.dropout_params:
                velocity = theano.shared(numpy.zeros(param.get_value().shape,
                                                     dtype = theano.config.floatX)
                                        )
                self.param_updates.append(velocity)
        
        # compute the gradients with respect to the model parameters inc reg.
        #gparams = T.grad(self.finetune_cost+W_decay/2*self.L2_reg, self.params)
        
        gparams = T.grad(self.dropout_cost+W_decay/2*self.L2_reg, self.dropout_params)
        # compute list of fine-tuning updates including momentum (weird implementation)
        updates= []
        
        if mom_method == 'classic':
            parcount = 0
            for param, prev_update, grad in zip(self.dropout_params,self.param_updates,gparams):
                #find updateterm
                
                updateterm = momentum*prev_update-learning_rate*grad
                # for weight matrices use max norm regularization
                if parcount%2 == 0 and MNreg:
                    new_param = self.max_norm_update(param+updateterm,c)
                else:
                    new_param = param+updateterm
           
                # add update expression for parameters
                updates.append((param,new_param))
            
                # add update expression for previous update/velocity:
                updates.append((prev_update, updateterm))
                
                parcount += 1
                
        elif mom_method == 'tutorial':
            parcount = 0
            for param, prev_update, grad in zip(self.dropout_params,self.param_updates,gparams):
                #find updateterm
                updateterm = (momentum*prev_update+(1.-momentum)*grad)
                
                # for weight matrices use max norm regularization
                if parcount%2 == 0 and MNreg:
                    new_param = self.max_norm_update(param-learning_rate*updateterm,c)
                else:
                    new_param = param-learning_rate*updateterm
           
                # add update expression for parameters
                updates.append((param,new_param))
            
                # add update expression for previous update/velocity:
                updates.append((prev_update, updateterm))
                
                parcount += 1
 

        
        elif mom_method == 'rmsprop':
            parcount = 0
            for param, prev_update, grad in zip(self.dropout_params,self.param_updates,gparams):

                
                # for weight matrices use max norm regularization
                if parcount%2 == 0 and MNreg:
                    new_param = self.max_norm_update(
                        param-T.switch(learning_rate/ T.sqrt(prev_update+1e-7)<5.,
                                       learning_rate/ T.sqrt(prev_update+1e-7),
                                       5.)*grad,c
                    )
                else:
                    new_param = param-T.switch((learning_rate+momentum)/ T.sqrt(prev_update+1e-7)<5.,
                                               (learning_rate+momentum)/ T.sqrt(prev_update+1e-7),
                                               5.)*grad 
           
                # add update expression for parameters
                updates.append((param,new_param))
            
                # add update expression for previous update/velocity:
                updates.append((prev_update, 0.9*prev_update+0.1*grad**2))
                
                parcount += 1
                
        else: # 'weird'
            parcount = 0
            for param, prev_update, grad in zip(self.dropout_params,self.param_updates,gparams):
                #find updateterm
                updateterm = learning_rate*(momentum*prev_update-(1.-momentum)*grad)
                
                # for weight matrices use max norm regularization
                if parcount%2 == 0 and MNreg:
                    new_param = self.max_norm_update(param+updateterm,c)
                else:
                    new_param = param+updateterm
           
                # add update expression for parameters
                updates.append((param,new_param))
            
                # add update expression for previous update/velocity:
                updates.append((prev_update, updateterm))
                
                parcount += 1
    
        
#        
#        for i in range(len(self.params)):
#           #self.param_updates[i] = momentum*self.param_updates[i]-learning_rate*gparams[i]

#            # update parameters
#            updates.append((self.params[i],self.params[i]-learning_rate*gparams[i]+momentum*self.param_updates[i]))
#            
#            updates.append((self.param_updates[i],momentum*self.param_updates[i]-learning_rate*gparams[i]))
        
                           
        
        print(updates)                  

        train_fn = theano.function(
            inputs=[index,momentum,learning_rate],
            outputs=[self.dropout_cost,self.finetune_cost],
            updates=updates,
            givens={
                self.x: train_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: train_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )

        test_score_i = theano.function(
            [index],
            outputs=self.linLayer.MSE(self.y),
            givens={
                self.x: test_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: test_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )

        valid_score_i = theano.function(
            [index],
            outputs=self.linLayer.MSE(self.y),
            givens={
                self.x: valid_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: valid_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )
        
        train_score_i = theano.function(
            [index],
            outputs=self.linLayer.MSE(self.y),
            givens={
                self.x: train_set_x[
                    index * batch_size: (index + 1) * batch_size
                ],
                self.y: train_set_y[
                    index * batch_size: (index + 1) * batch_size
                ]
            }
        )
        
        # Create a function that scans the entire validation set
        def valid_score():
            return [valid_score_i(i) for i in xrange(n_valid_batches)]

        # Create a function that scans the entire test set
        def test_score():
            return [test_score_i(i) for i in xrange(n_test_batches)]
        
        # Create a function that scans the entire training set
        def train_score():
            return [train_score_i(i) for i in xrange(n_train_batches)]

        return train_fn, valid_score, test_score, train_score


def _dropout_from_layer(rng, layer, p,dist):
    """p is the probablity of dropping a unit
    """
    srng = theano.tensor.shared_randomstreams.RandomStreams(
            rng.randint(999999))
    # p=1-p because 1's indicate keep and p is prob of dropping
    if dist =='normal':
        mask = srng.normal(size=layer.shape,avg = 1.,std = numpy.sqrt(p/(1-p)))
    else:
        mask = srng.binomial(n=1, p=1-p, size=layer.shape)
    # The cast is important because
    # int * float32 = float64 which pulls things off the gpu
    output = layer * T.cast(mask, theano.config.floatX)
    return output

class DropoutHiddenLayer(HiddenLayer):
    def __init__(self, rng, input, n_in, n_out,
                 activation, dropout_rate,dist = 'binomial', W=None, b=None,normal_init=False):
        super(DropoutHiddenLayer, self).__init__(
                rng=rng, input=input, n_in=n_in, n_out=n_out, W=W, b=b,
                activation=activation,normal_init=normal_init)

        self.output = _dropout_from_layer(rng, self.output, p=dropout_rate,
                                         dist = dist)
        

def train_DBN(finetune_lr=0.0005, training_epochs=100,
             batch_size=128,finetune_wd = 0.,
             momentum = 0.5,momentum_method = 'tutorial',
             use_MN = True,max_norm=2.6,use_pretrained_params=False):
    """
    Demonstrates how to train and test a Deep Belief Network.

    :type finetune_lr: float
    :param finetune_lr: learning rate used in the finetune stage
    :type pretraining_epochs: int
    :param pretraining_epochs: number of epoch to do pretraining
    :type pretrain_lr: float
    :param pretrain_lr: learning rate to be used during pre-training
    :type k: int
    :param k: number of Gibbs steps in CD/PCD
    :type training_epochs: int
    :param training_epochs: maximal number of iterations ot run the optimizer
    :type dataset: string
    :param dataset: path the the pickled dataset
    :type batch_size: int
    :param batch_size: the size of a minibatch
    :type finetune_wd: float
    :param finetune_wd: weight decay (L2_reg) coefficient
    :type momentum: float
    :param momentum: 0<= momentum <1 
    """
    writedict ={}
    Nfrq = 129
    left_cfr = 2
    right_cfr = 2
    numb_cfr = left_cfr + 1 + right_cfr 
    n_ins = numb_cfr*Nfrq
    SNRlist = [0,5,10]
    
    numb_train_fr = 4*224896 # 224896 frames is about 1 hour
    numb_valid_fr = 1*74880
    numb_test_fr =1*74880
    
    dropout_dist = 'normal'
    
    noisetypes = ['AWGN','babble','restaurant','street']
    validnoise = ['AWGN','babble','restaurant','street'] # added one unseen case to validation
    testnoise = ['car','exhibition']
    
    #loading data:
    print('loading data...')
    train_load_x, train_load_y,next_line_tr = load_data_2(filelist = 'train_stem_big.list',
                                             path = 'training_preprosd/',
                                             Nframes = numb_train_fr,
                                             left_cfr = left_cfr,
                                             right_cfr = right_cfr,
                                             Nfrq = Nfrq,
                                             SNRlist = SNRlist,
                                             noisetypes = noisetypes,
                                             shuffle_whole = True,
                                             start_line = 0)
    
    print('read %i lines from TIMIT for training set' % (next_line_tr-1))
                                         
    valid_load_x, valid_load_y,next_line_val = load_data_2(filelist = 'test_overlap_stem_big.list',
                                             path = 'validation_preprosd/',
                                             Nframes = numb_valid_fr,
                                             left_cfr = left_cfr,
                                             right_cfr = right_cfr,
                                             Nfrq = Nfrq,
                                             SNRlist = SNRlist,
                                             noisetypes = validnoise,
                                             shuffle_whole = True,
                                             start_line = 0)
    
    print('read %i lines from TIMIT for validation set' % (next_line_val-1))
                                         
    test_load_x, test_load_y,next_line_test = load_data_2(filelist = 'test_separate_stem.list',
                                           path = 'test_preprosd/',
                                           Nframes = numb_test_fr,
                                           left_cfr = left_cfr,
                                           right_cfr = right_cfr,
                                           Nfrq = Nfrq,
                                           SNRlist = SNRlist,
                                           noisetypes = testnoise,
                                           shuffle_whole = True,
                                           start_line = 0)
    
    print('read %i lines from TIMIT for test set' % (next_line_test-1))
    
    traindat_sec = train_load_x.shape[0]*(16/1000.)
    traindat_min = traindat_sec/60.
    traindat_hou = int(traindat_min/60.)
    traindat_min_rest = int(traindat_min-traindat_hou*60.)
    traindat_sec_rest = int(traindat_sec-int(traindat_min)*60.)
   
    
    print('training data is %i number of frames, equaling %d mini-batches' 
          % (train_load_x.shape[0],train_load_x.shape[0]/128.))
    print('This euqals about %i hours, %i minutes and %i seconds of speech data' 
          % (traindat_hou,traindat_min_rest,traindat_sec_rest))
    
    #normalize data (keep normalization factors)
    
    datamean_in = numpy.mean(train_load_x, axis = 0)
    datastd_in = numpy.std(train_load_x, axis = 0,ddof=1)
    
    datamean_out = datamean_in[(numb_cfr-1)/2*Nfrq:((numb_cfr-1)/2+1)*Nfrq]
    datastd_out = datastd_in[(numb_cfr-1)/2*Nfrq:((numb_cfr-1)/2+1)*Nfrq]
    
    if not (datastd_in==0).sum():
        print('scaling data...')
        for i in range(numb_cfr*Nfrq):
            train_load_x[:][:,i] = (train_load_x[:][:,i]-datamean_in[i]) / datastd_in[i]
            valid_load_x[:][:,i] = (valid_load_x[:][:,i]-datamean_in[i]) / datastd_in[i]
            test_load_x[:][:,i] = (test_load_x[:][:,i]-datamean_in[i]) / datastd_in[i]
                
        for i in range(Nfrq):
            train_load_y[:][:,i] = (train_load_y[:][:,i]-datamean_out[i]) / datastd_out[i]
            valid_load_y[:][:,i] = (valid_load_y[:][:,i]-datamean_out[i]) / datastd_out[i]
            test_load_y[:][:,i] = (test_load_y[:][:,i]-datamean_out[i]) / datastd_out[i]
    else:
        sys.exit('Error! Zero in feature scaling factor')
    

    
    #store data in theano.shared variables
    print('storing data in theano.shared variables...')
    train_set_x = theano.shared(train_load_x,borrow=True)
    valid_set_x = theano.shared(valid_load_x,borrow=True)
    test_set_x = theano.shared(test_load_x,borrow=True)
                                       
    train_set_y = theano.shared(train_load_y,borrow=True)
    valid_set_y = theano.shared(valid_load_y,borrow=True)
    test_set_y = theano.shared(test_load_y,borrow=True)
    
    #define list of datasets as used in example code
    datasets = [(train_set_x, train_set_y), (valid_set_x, valid_set_y),
                (test_set_x, test_set_y)]
                
    # compute number of minibatches for training, validation and testing
    n_train_batches = train_set_x.get_value(borrow=True).shape[0] / batch_size
    


    # numpy random generator
    numpy_rng = numpy.random.RandomState(5098)
    print('... building the model')
    
    # construct the Deep Belief Network
 
    hidden_sizes = [500,500,500]
    dropout_rates=[0.1, 0.2, 0.2,0.2]
    
    
    dbn = DNN(numpy_rng=numpy_rng, n_ins=n_ins,
              hidden_layers_sizes=hidden_sizes,
              n_outs=129,dropout_rates=dropout_rates,
              normal_init = False, dropout_dist =dropout_dist )
    
    # for storing best NN given validation set performance
    dbn_best = DNN(numpy_rng=numpy_rng, n_ins=n_ins,
              hidden_layers_sizes=hidden_sizes,
              n_outs=129,dropout_rates=dropout_rates, 
              dropout_dist =dropout_dist)

    # for storing best NN given training set performance
    # to be used for resetting momentum if "rolling past"
    # minimum when using momentum
    dbn_best_training = DNN(numpy_rng=numpy_rng, n_ins=n_ins,
              hidden_layers_sizes=hidden_sizes,
              n_outs=129,dropout_rates=dropout_rates, 
              dropout_dist =dropout_dist)
                
    
    ########################
    # FINETUNING THE MODEL #
    ########################
    

    # get the training, validation and testing function for the model
    print('... getting the finetuning functions')
    train_fn, validate_model, test_model, train_performance = dbn.build_finetune_functions(
        datasets=datasets,
        batch_size=batch_size,
        W_decay = finetune_wd,
        mom_method = momentum_method,
        c=max_norm,
        MNreg=use_MN
    )

    print('... finetuning the model')
    # early-stopping parameters
    patience = 4 * n_train_batches  # look as this many examples regardless
    patience_increase = 10.    # wait this much longer when a new best is 
                              # found
    improvement_threshold = 0.995  # a relative improvement of this much is
                                   # considered significant
    validation_frequency = min(n_train_batches, patience / 2)
                                  # go through this many
                                  # minibatches before checking the network
                                  # on the validation set; in this case we
                                  # check every epoch

    best_validation_loss = numpy.inf
    best_training_loss = numpy.inf
    best_params_testscore = numpy.inf
    test_score = 0.
    start_time = timeit.default_timer()

    done_looping = False
    epoch = 0
    best_iter = 0
    
    #minibatch_cost_dev = []
    training_cost_dev = []
    validation_cost_dev = []
    test_cost_dev = []
    
    momentum_dev = []
    maxmom = 0.9
    mom_inc = True
    
    FT_lr =finetune_lr
    lr_scale = 0.99
    dtype = numpy.dtype(theano.config.floatX).type
    
    #final_layer_norms = []
    
    rmsprop_lrmean = []
    rmsprop_lrmax = []
    rmsprop_lrmin = []
    while (epoch < training_epochs) and (not done_looping):
        
        
        # supervise norms of incoming weight vectors to firs layer
        #E_w = numpy.transpose(numpy.sqrt(numpy.sum(dbn.params[0].get_value()**2,axis=0,keepdims=True)))
        #final_layer_norms.append(E_w)
        
        if epoch == 0:
            training_losses = train_performance()
            this_training_loss = numpy.mean(training_losses)
            training_cost_dev.append(this_training_loss)

            test_losses = test_model()
            test_score = numpy.mean(test_losses)
            test_cost_dev.append(test_score)

            validation_losses = validate_model()
            this_validation_loss = numpy.mean(validation_losses)
            validation_cost_dev.append(this_validation_loss)
            print(' Initial training loss %s, validation loss %s, and test loss %s' 
                  % (
                    this_training_loss,
                    this_validation_loss,
                    test_score)
                  )
            
        
        epoch = epoch + 1
        
        for minibatch_index in xrange(n_train_batches):
            
            minibatch_avg_cost = train_fn(minibatch_index,momentum,FT_lr)
            #minibatch_cost_dev.append(minibatch_avg_cost)
            iter = (epoch - 1) * n_train_batches + minibatch_index

#            momentumval = numpy.minimum(1-1/(2*(numpy.floor(iter/250)+1)),maxmom)
#            momentum = numpy.asarray(momentumval,dtype = theano.config.floatX)
            if (iter + 1) % validation_frequency == 0:

                validation_losses = validate_model()
                this_validation_loss = numpy.mean(validation_losses)
                validation_cost_dev.append(this_validation_loss)
                print(
                    'epoch %i, minibatch %i/%i, validation cost  %f'
                    % (
                        epoch,
                        minibatch_index + 1,
                        n_train_batches,
                        this_validation_loss
                    )
                )
                
#                if minibatch_index==0:
#                    done_looping = True
#                    break
                # test it on the test set
                test_losses = test_model()
                test_score = numpy.mean(test_losses)
                test_cost_dev.append(test_score)
            
                # if we got the best validation score until now
                if this_validation_loss < best_validation_loss:

                    #improve patience if loss improvement is good enough
                    if (
                        this_validation_loss < best_validation_loss *
                        improvement_threshold
                    ):
                        patience = max(patience, iter * patience_increase)

                    # save best validation score and iteration number
                    best_validation_loss = this_validation_loss
                    best_iter = iter
                    best_params_testscore = test_score
                    
                    # store parameter values in dbn_best
                    for paramnumb in range(len(dbn.params)):
                        dbn_best.dropout_params[paramnumb].set_value(dbn.dropout_params[paramnumb].get_value(borrow=True))
                    
                        
                    print(('     epoch %i, minibatch %i/%i, test cost of '
                           'best model %f') %
                          (epoch, minibatch_index + 1, n_train_batches,
                           test_score))
             
                
#            if patience <= iter:
#                done_looping = True
#                break
        if momentum_method =='rmsprop':
            rmsprop_lr = FT_lr/numpy.sqrt(dbn.param_updates[0].get_value())
            rmsprop_lrmean.append(numpy.mean(rmsprop_lr))
            rmsprop_lrmax.append(numpy.amax(rmsprop_lr))
            rmsprop_lrmin.append(numpy.amin(rmsprop_lr))
        #increase momentum
        if mom_inc and epoch>5 and momentum<=maxmom:
            momentum += 0.01
            if momentum<0.5:
                momentum = 0.5
                
        if momentum>=maxmom: #after momentum is max, start scaling learning rate
            FT_lr = dtype(FT_lr*lr_scale)
        #if epoch>int(0.8*training_epochs) and momentum>=0.9:
        #    momentum -= 0.005
                    
        # find full-batch training set MSE
        training_losses = train_performance()
        this_training_loss = numpy.mean(training_losses)
        training_cost_dev.append(this_training_loss)
                
        #store best training loss
        if this_training_loss < best_training_loss*improvement_threshold:
            best_training_loss = this_training_loss
            for paramnumb in range(len(dbn.params)):
                dbn_best_training.dropout_params[paramnumb].set_value(dbn.dropout_params[paramnumb].get_value(borrow=True))
        else: 
            # if training cost is 5 times the lowest so far, reset to best parameters
            # and set momentum to 0
            if this_training_loss/best_training_loss > 5.:
                # set momentum to clear history in velocity term 
                momentum = 0.
                        
                # updating parameters to equal best performance on training set so far
                for paramnumb in range(len(dbn.params)):
                    dbn.dropout_params[paramnumb].set_value(dbn_best_training.dropout_params[paramnumb].get_value(borrow=True))
                
                
        momentum_dev.append(momentum)
        
    end_time = timeit.default_timer()
    
    
    
    print(
        (
            'Optimization complete with best validation score of %f %%, '
            'obtained at iteration %i, '
            'with test performance %f'
        ) % (best_validation_loss, best_iter + 1, best_params_testscore)
    )
    print >> sys.stderr, ('The fine tuning code for file ' +
                          os.path.split(__file__)[1] +
                          ' ran for %.2fm' % ((end_time - start_time)
                                              / 60.))
    
    test_set_x.set_value([[]])
    test_set_y.set_value([[]])
    valid_set_x.set_value([[]])
    valid_set_y.set_value([[]])
    
    GVref, GVref_const , GVest , GVest_const= dbn_best.GV_est(train_set_x[0:int(train_set_x.get_value().shape[0]/2)],
                                                              train_set_y[0:int(train_set_x.get_value().shape[0]/2)])
    GVeq = numpy.sqrt(GVref/GVest)
    GVeq_const = numpy.sqrt(GVref_const/GVest_const)
    
    #avoid scaling weights if gaussian dropout is used
    
    w_scale = []
    for p in dropout_rates:
        if dropout_dist == 'normal':
            w_scale.append(p*0.)
        else:
            w_scale.append(p)
    
    matlabs = {} #for saving dbn.params for matlab
    matlabs2 = {} #for saving dbn_best.params for matlab
    
    # save parameters
    save_params = open('parameters/dbn.save','wb')
    cPickle.dump(dbn,save_params,-1)
    cPickle.dump(datamean_in,save_params,-1)
    cPickle.dump(datastd_in,save_params,-1)
    paramtype = ['W','b']
    
    matlabs['my_in'] = datamean_in
    matlabs['sigma_in'] = datastd_in
    matlabs['my_tar'] = datamean_out
    matlabs['sigma_tar'] = datastd_out
    
    matlabs2['my_in'] = datamean_in
    matlabs2['sigma_in'] = datastd_in
    matlabs2['my_tar'] = datamean_out
    matlabs2['sigma_tar'] = datastd_out
   
    matlabs2['training_dev'] = training_cost_dev
    matlabs2['validation_dev'] = validation_cost_dev
    matlabs2['test_dev'] = test_cost_dev
    matlabs2['momentum_dev'] = momentum_dev
    matlabs2['GVref'] = GVref
    matlabs2['GVest'] = GVest
    matlabs2['GVeq'] = GVeq
    matlabs2['GVref_const'] = GVref_const
    matlabs2['GVest_const'] = GVest_const
    matlabs2['GVeq_const'] = GVeq_const
    matlabs2['RMSprop_LRmean_W0'] = rmsprop_lrmean
    matlabs2['RMSprop_LRmax_W0'] = rmsprop_lrmax
    matlabs2['RMSprop_LRmin_W0'] = rmsprop_lrmin
    #matlabs2['norm_dev'] = final_layer_norms
    for i in range(len(dbn.params)):
        if i%2==0: #parameter is weight matrix
            matlabs[paramtype[i%2]+str(int(i/2))]=dbn.dropout_params[i].get_value(borrow=True)*(1 - w_scale[int(i/2)])
            matlabs2[paramtype[i%2]+str(int(i/2))]=dbn_best.dropout_params[i].get_value(borrow=True)*(1 - w_scale[int(i/2)])
        else:
            matlabs[paramtype[i%2]+str(int(i/2))]=dbn.dropout_params[i].get_value(borrow=True)
            matlabs2[paramtype[i%2]+str(int(i/2))]=dbn_best.dropout_params[i].get_value(borrow=True)
    

        
    
    
    
    save_params.close()
    scipy.io.savemat('parameters/ReLU_exp.mat',matlabs)
    scipy.io.savemat('parameters/ReLU_exp_best.mat',matlabs2)
    
    # storing metadata
    metadata = []
    
    metadata.append('FT_lr = ' + str(finetune_lr) + '\n')
    metadata.append('FT_epochs = ' + str(training_epochs) + '\n')
    metadata.append('momentum_method = ' + str(momentum_method) + '\n')
    metadata.append('use_MN = ' + str(use_MN) + ' with max_norm ' + str(max_norm)+'\n')
    
    metadata.append('number of training frames = ' + str(numb_train_fr) + ' = about ' +str(numb_train_fr/224896.) +  ' hours\n')
    metadata.append('number of context frames in input, left = ' + str(left_cfr) + ' right = ' + str(right_cfr) + '\n')
    metadata.append('number of TIMIT files used for training = ' + str(next_line_tr) + '\n')
    metadata.append('number of TIMIT files used for validation = ' + str(next_line_val) + '\n')
    metadata.append('number of TIMIT files used for test = ' + str(next_line_test) + '\n')
    metadata.append('noise types in training set = ' + str(noisetypes) +'\n')
    metadata.append('noise types in validation set = ' + str(validnoise) +'\n')
    metadata.append('test noise types= ' + str(testnoise)+ '\n')
    metadata.append('SNR levels= ' + str(SNRlist) +'\n')
    metadata.append('hidden layer sizes = ' + str(hidden_sizes) +'\n')
    metadata.append('dropout rates = ' + str(dropout_rates) +'\n')
    metadata.append('dropout distibution = ' + str(dropout_dist) +'\n')
    metadata.append('learning rate scaling= ' + str(lr_scale)+ '\n')
    
    metadatafile = open('parameters/DNN_ReLU_metadata.txt','w')
    for metaline in metadata:
        metadatafile.write(metaline)
        
    metadatafile.close()
    
    
    #n = NN(int(len(dbn.params)/2),dbn.params,datamean_in,datastd_in,datamean_out,datastd_out)
    #save_nn = open('NNtrained.save','wb')
    #cPickle.dump(nn,save_nn,-1)
    #save_nn.close()
   
    



if __name__ == '__main__':
    train_DBN()